from meshagent.api.schema import MeshSchema, ElementType, ChildProperty, ValueProperty

presentation_schema = MeshSchema(
    root_tag_name="presentation",
    elements=[  
        ElementType(
            tag_name="presentation", 
            description="a presentation",
            properties=[
                ChildProperty(name="slides", description="the slides for the presentation", child_tag_names=["slide"]),
                ]
        ),
        ElementType(
            tag_name="slide", 
            description="a slide",
            properties=[
                ValueProperty(name="title", description="a title for the slide", type="string"),
                ValueProperty(name="background", description="an file to use for an background of the slide, must be a filename output from a previous step", type="string"),
                ChildProperty(name="lines", description="the slides for the presentation", child_tag_names=["line"]),
                ]
        ),

            ElementType(
            tag_name="line", 
            description="a line of text",
            properties=[
                ValueProperty(name="line", description="the text of a bullet point for the slide", type="string"),
            ]
            )
            
    ]
)




from .adapter import LLMAdapter, Toolkit, ToolResponseAdapter
from meshagent.api.schema_util import prompt_schema
from .agent import AgentCallContext
from typing import Optional
from meshagent.agents import TaskRunner
from meshagent.api import RequiredToolkit

# An agent that takes a simple prompt and gets the result
class PromptAgent(TaskRunner):
    def __init__(self,
        *,
        name: str,
        output_schema: dict,
        llm_adapter: LLMAdapter,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        tools: list[Toolkit] = [],
        rules: list[str] = [],
        title: Optional[str] = None,
        description: Optional[str] = None,
        requires: Optional[list[RequiredToolkit]] = None,
        supports_tools: Optional[bool] = None,
        labels: Optional[list[str]] = None
    ):
        super().__init__(
            name=name,
            description=description,
            title=title,
            input_schema=prompt_schema(
                description=description
            ),
            output_schema=output_schema,
            requires=requires,
            supports_tools=supports_tools,
            labels=labels
        )
        self.rules = rules
        self.tools = tools
        self.llm_adapter = llm_adapter
        self.tool_adapter = tool_adapter

    async def init_chat_context(self):
        chat = self.llm_adapter.create_chat_context()
        chat.append_rules(self.rules)
        return chat

    async def ask(self, *, context: AgentCallContext, arguments: dict):
        context.chat.append_user_message(arguments["prompt"])

        toolkits = [
            *self.toolkits,
            *context.toolkits
        ]

        return await self.llm_adapter.next(context=context.chat, room=context.room, toolkits=toolkits, tool_adapter=self.tool_adapter, output_schema=self.output_schema)


from .agent import TaskRunner, AgentCallContext
from meshagent.api.chan import Chan
from meshagent.api import RoomMessage, RoomException, RoomClient, RemoteParticipant
from meshagent.tools import Toolkit
from .adapter import LLMAdapter, ToolResponseAdapter
import asyncio
from typing import Optional
import json

import logging

logger = logging.getLogger("chat")



# todo: thread should stop when participant stops?

class Worker(TaskRunner):
    def __init__(self, *, queue: str, prompt: str,  name, title = None, description = None, requires = None, llm_adapter: LLMAdapter, tool_adapter:  Optional[ToolResponseAdapter] = None, toolkits: Optional[list[Toolkit]] = None, rules : Optional[list[str]] = None, supports_tools: bool = True):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            output_schema=None,
            supports_tools=supports_tools
        )

        self._queue = queue
        self._prompt = prompt

        if toolkits == None:
            toolkits = []

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter

        self._message_channel = Chan[RoomMessage]()

        self._room : RoomClient | None = None
        self._toolkits = toolkits

        if rules == None:
            rules = []

        self._rules = rules


    async def ask(self, *, context: AgentCallContext, arguments: dict):
        
        queue = self._queue
        prompt = self._prompt

        step_schema = {
            "type" : "object",
            "required" : ["text","finished"],
            "additionalProperties" : False, 
            "description" : "execute a step",
            "properties" : {
                "text" : {
                    "description" : "a reply to the user or status to display during an intermediate step",
                    "type" : "string"
                },
                "finished" : {
                    "description" : "whether the agent has finished answering the user's last message, also should be set to true if we get stuck in a loop, or if the user did not make a request",
                    "type" : "boolean"
                }
            }
        }
        
        # todo: add graceful exit 

        while True:

            message = await self.room.queues.receive(name=queue, create=True, wait=True)
            if message != None:
                
                # for each message, create a new chat context

                chat_context = await self.init_chat_context()

            
                chat_context.append_rules(
                    rules=[
                        *self._rules,
                    ]
                )
                
                chat_context.append_user_message(message=prompt)
                chat_context.append_user_message(message=json.dumps(message))
                
                try:
                    while True:

                        tool_target = context.caller
                        if context.on_behalf_of != None:
                            tool_target = context.on_behalf_of

                        response = await self._llm_adapter.next(
                            context=chat_context,
                            room=self._room,
                            toolkits=context.toolkits,
                            tool_adapter=self._tool_adapter,
                            output_schema=step_schema,
                        )

                        if response["finished"] or len(context.toolkits) == 0:
                            break
                        else:
                            chat_context.append_user_message(message="proceed to the next step if you are ready")
                    
                except Exception as e:

                    logger.error(f"Failed to process a message {message}", exc_info=e)



from pydantic import BaseModel
from abc import abstractmethod

import pydantic_ai.models
from .agent import AgentCallContext
from .writer import Writer, WriterContext
import logging
from typing import Optional
from meshagent.api import RoomClient

from .agent import TaskRunner, RequiredToolkit, Requirement, RequiredSchema

from .adapter import  ToolResponseAdapter
from meshagent.tools.toolkit import Tool, Response, ToolContext
from meshagent.tools.pydantic import get_pydantic_ai_tool_definition

from typing import Sequence
import pydantic_ai

logger = logging.getLogger("pydantic_agent")



def get_pydantic_ai_tool(*, room: RoomClient, tool: Tool, response_adapter: ToolResponseAdapter) -> pydantic_ai.tools.Tool:
     async def prepare(ctx: pydantic_ai.RunContext, tool_def: pydantic_ai.tools.ToolDefinition):
         return get_pydantic_ai_tool_definition(tool=tool)
     
     async def execute(**kwargs):
         response = await tool.execute(context=ToolContext(room=room, caller=room.local_participant), **kwargs)
         return await response_adapter.to_plain_text(room=room, response=response)

     return pydantic_ai.Tool(
        name=tool.name,
        takes_ctx=False,
        description=tool.description,
        prepare=prepare,
        function=execute
    )

def get_pydantic_ai_tools_from_context(*, context: AgentCallContext, response_adapter: ToolResponseAdapter) -> list[pydantic_ai.tools.Tool]:

    tools = list[pydantic_ai.tools.Tool]()

    for toolkit in context.toolkits:
        
        for tool in toolkit.tools:

            tools.append(get_pydantic_ai_tool(room=context.room, tool=tool, response_adapter=response_adapter))

    return tools

class PydanticAgent[TInput:BaseModel, TOutput:BaseModel](TaskRunner):
    def __init__(self,
        *,
        name: str,
        input_model: TInput,
        output_model: TOutput,
        title: Optional[str] = None,
        description: Optional[str] = None,
        requires: Optional[list[Requirement]] = None,
        supports_tools : Optional[bool] = None,
    ):
        self.input_model = input_model
        self.output_model = output_model
        
        super().__init__(
            name = name,
            description = description,
            title = title,
            input_schema=input_model.model_json_schema(),
            output_schema=output_model.model_json_schema(),   
            requires=requires,
            supports_tools=supports_tools
        )
    
    async def ask(self, *, context: AgentCallContext, arguments: dict):
        try:
            input = self.input_model.model_validate(arguments)
            output = await self.ask_model(context=context, arguments=input)
            return output.model_dump(mode='json')
        except Exception as e:
            logger.error("Unhandled exception in ask agent call", exc_info=e)
            raise

    @abstractmethod
    async def ask_model(self, context: AgentCallContext, arguments: TInput) -> TOutput:
        pass

class PydanticWriter[TInput:BaseModel, TOutput:BaseModel](Writer):
    def __init__(self, 
        name: str,
        input_model: TInput,
        output_model: TOutput,
        title: Optional[str] = None,
        description: Optional[str] = None
    ):

        self.input_model = input_model
        self.output_model = output_model

        super().__init__(           
            name=name,
            description=description,
            title=title, 
            input_schema=input_model.model_json_schema(),
            output_schema=output_model.model_json_schema(),
        )

    async def write(self, writer_context: WriterContext, arguments: dict):
        try:
            input = self.input_model.model_validate(arguments)
            output = await self.write_model(writer_context=writer_context, arguments=input)
            return output.model_dump(mode='json')
        except Exception as e:
            logger.error("Unhandled exception in write call", exc_info=e)
            raise
    
    @abstractmethod
    async def write_model(self, writer_context: WriterContext, arguments: TInput) -> TOutput:
        pass



class PydanicAgentHost[TInput:BaseModel, TOutput:BaseModel](PydanticAgent[TInput, TOutput]):
    
    def __init__(self, *, 
                model: pydantic_ai.models.KnownModelName, 
                system_prompt: str | Sequence[str],
                response_adapter: ToolResponseAdapter, 
                tools: Optional[list[pydantic_ai.Tool]] = None,
                name, input_model, output_model, title = None, description = None, requires = None, supports_tools = None):
        super().__init__(
            name=name, input_model=input_model, output_model=output_model, title=title, description=description, requires=requires, supports_tools=supports_tools, 
            
        )
        self._model = model
        self._system_prompt = system_prompt
        self._response_adapter = response_adapter
        if tools == None: 
            tools = []
            
        self._tools = tools

    def message_history(self, context: AgentCallContext) -> None | list:
        return None
    
    def to_prompt(self, *, arguments: TInput) -> str:
        return str(arguments)

    def create_agent(self, *, tools: list[pydantic_ai.Tool]) -> pydantic_ai.Agent[TInput,TOutput]:
        return pydantic_ai.Agent( 
            model=self._model,
            system_prompt=self._system_prompt,
            deps_type=self.input_model,
            result_type=self.output_model,
            tools=[
                *self._tools,
                *tools
            ]
        )

    @abstractmethod
    async def ask_model(self, context: AgentCallContext, arguments: TInput) -> TOutput:
        
        agent = self.create_agent(tools=get_pydantic_ai_tools_from_context(context=context, response_adapter=self._response_adapter))
        result = await agent.run(self.to_prompt(arguments=arguments), message_history=self.message_history(context), result_type=self.output_model, deps=arguments)
        return result.data


from meshagent.api.schema import MeshSchema, ElementType,  ChildProperty, ValueProperty

thread_schema = MeshSchema(
    root_tag_name="thread",
    elements=[
        ElementType(
            tag_name="thread", 
            description="a thread of messages",
            properties=[
                ValueProperty(name="name", description="name of the tread", type="string"),
                ChildProperty(name="properties", description="the messages in the thread", ordered=True, child_tag_names=[
                    "members", "messages"
                ]),
            ]
        ),
        ElementType(
            tag_name="members", 
            description="the members of this thread",
            properties=[
                ChildProperty(name="items", child_tag_names=["member"], description="the messages in this thread")
            ]
        ),
        ElementType(
            tag_name="messages", 
            description="the messages of this thread",
            properties=[
                ChildProperty(name="items", child_tag_names=["message"], description="the messages in this thread")
            ]
        ),
        ElementType(
            tag_name="member", 
            description="a member of this thread",
            properties=[
                ValueProperty(name="name", description="the name of the member", type="string"),
                ValueProperty(name="type", description="the type of member", type="string", enum=[
                    "user", "agent"
                ]),
            ]
        ),
        ElementType(
            tag_name="file", 
            description="a file attachment",
            properties=[
                ValueProperty(name="path", description="the path of the file in the room", type="string"),
            ]
        ),
        ElementType(
            tag_name="message", 
            description="a message sent in the conversation",
            properties=[
                ValueProperty(name="id", description="the id of the message", type="string"),
                ValueProperty(name="text", description="the text of the message", type="string"),
                ValueProperty(name="created_at", description="the date that the message was sent in ISO format", type="string"),
                ValueProperty(name="author_name", description="the name of the author of the post", type="string"),
                ValueProperty(name="author_ref", description="a reference to author identity in another system", type="string"),
                ChildProperty(name="attachments", child_tag_names=["file"], description="a list of message attachments")
            ]
        ),          
    ]
)



from meshagent.api.schema import MeshSchema, ElementType,  ChildProperty, ValueProperty, SimpleValue

super_editor_document_schema = MeshSchema(root_tag_name='document', elements=[
    ElementType(
      tag_name='document',
      description='A SuperEditor document for use with Flutter apps',
      properties=[
        ChildProperty(name='children', child_tag_names=[
          'block',
          'horizontalRule',
          'image',
        ]),
      ],
    ),
    ElementType(
      tag_name='block',
      description='Block element - this can be a paragraph, list-item or tasks',
      properties=[
        ValueProperty(name='blockType', type="string"),
        ValueProperty(name='textAlign', type="string"),
        ValueProperty(name='itemType',  type="string"),
        ValueProperty(name='indent',  type="number"),
        ChildProperty(name='children', child_tag_names= ['text']),
      ],
    ),
    ElementType(
      tag_name='image',
      description='Image element',
      properties=[
        ValueProperty(name='imageUrl', type="string"),
        ValueProperty(name='width', type="number"),
        ValueProperty(name='height', type="number"),
      ],
    ),
    ElementType(
      tag_name='horizontalRule',
      description='Horizontal rule element',
      properties=[],
    ),
    ElementType(
      tag_name='text',
      description='Text element',
      properties=[
        ValueProperty(name='text', type="string"),
      ],
    ),
  ])


import logging

from meshagent.api import websocket_protocol, RoomMessage
from meshagent.api.webhooks import WebhookServer, RoomStartedEvent, RoomEndedEvent, WebSocketClientProtocol, CallEvent
from meshagent.api.room_server_client import RoomClient
from meshagent.agents import SingleRoomAgent
from aiohttp import web
import asyncio

import logging
from typing import Callable, Optional

from .agent import TaskRunner

logger = logging.getLogger("hosting")



class RemoteTaskRunnerServer[T:TaskRunner](WebhookServer):
    def __init__(self, *, cls: Optional[T] = None, path: Optional[str] = None, app: Optional[web.Application] = None, host = None, port = None, webhook_secret = None, create_agent: Optional[Callable[[dict],TaskRunner]] = None, validate_webhook_secret: Optional[bool] = None):
        super().__init__(path=path, app=app, host=host, port=port, webhook_secret=webhook_secret, validate_webhook_secret=validate_webhook_secret)
        
        if create_agent == None:
            def default_create_agent(arguments: dict) -> TaskRunner:
                return cls(**arguments)
            
            create_agent = default_create_agent

        self._create_agent = create_agent

    async def _spawn(self, *, room_name: str, room_url: str, token: str, arguments: Optional[dict] = None):
        agent = self._create_agent(arguments=arguments)
        
        async def run():
            async with RoomClient(protocol=WebSocketClientProtocol(url= room_url, token=token)) as room:
    
                dismissed = asyncio.Future()

                def on_message(message: RoomMessage):
                    if message.type == "dismiss":
                        logger.info(f"dismissed by {message.from_participant_id}")
                        dismissed.set_result(True)

                room.messaging.on("message", on_message)
        
                await agent.start(room=room)

                done, pending = await asyncio.wait([
                    dismissed,
                    asyncio.ensure_future(room.protocol.wait_for_close())
                ], return_when=asyncio.FIRST_COMPLETED)
                

                await agent.stop()

        def on_done(task: asyncio.Task):
            try:
                result = task.result()
            except Exception as e:
                logger.error("agent encountered an error", exc_info=e)
                
        task = asyncio.create_task(run())
        task.add_done_callback(on_done)
    
    async def on_call(self, event: CallEvent):
        await self._spawn(room_name=event.room_name, room_url=event.room_url, token=event.token, arguments=event.arguments)


class RemoteAgentServer[T:SingleRoomAgent](WebhookServer):
    def __init__(self, *, cls: Optional[T] = None, path: Optional[str] = None, app: Optional[web.Application] = None,  host = None, port = None, webhook_secret = None, create_agent: Optional[Callable[[dict], SingleRoomAgent]] = None, validate_webhook_secret: Optional[bool] = None):
        super().__init__(path=path, app=app, host=host, port=port, webhook_secret=webhook_secret, validate_webhook_secret=validate_webhook_secret)
        
        if create_agent == None:
            def default_create_agent(arguments: dict) -> SingleRoomAgent:
                return cls(**arguments)
            
            create_agent = default_create_agent

        self._create_agent = create_agent
    
    async def _spawn(self, *, room_name: str, room_url: str, token: str, arguments: Optional[dict] = None):

        logger.info(f"room: {room_name} url: {room_url} token: {token} arguments: {arguments}")
        agent = self._create_agent(arguments=arguments)

        async def run():
            async with RoomClient(protocol=WebSocketClientProtocol(url=room_url, token=token)) as room:
    
                dismissed = asyncio.Future()

                def on_message(message: RoomMessage):
                    if message.type == "dismiss":
                        logger.info(f"dismissed by {message.from_participant_id}")
                        dismissed.set_result(True)

                room.messaging.on("message", on_message)
            
                await agent.start(room=room)

                done, pending = await asyncio.wait([
                    dismissed,
                    asyncio.ensure_future(room.protocol.wait_for_close())
                ], return_when=asyncio.FIRST_COMPLETED)
                
                await agent.stop()

        def on_done(task: asyncio.Task):
            try:
                result = task.result()
            except Exception as e:
                logger.error("agent encountered an error", exc_info=e)

        task = asyncio.create_task(run())
        task.add_done_callback(on_done)
        
    async def on_call(self, event: CallEvent):
        await self._spawn(room_name=event.room_name, room_url=event.room_url, token=event.token, arguments=event.arguments)




from .agent import Agent, AgentCallContext, AgentChatContext, RequiredToolkit, TaskRunner, SingleRoomAgent
from .development import connect_development_agent
from .listener import Listener, ListenerContext
from .hosting import RemoteTaskRunnerServer
from .adapter import ToolResponseAdapter, LLMAdapter
from .thread_schema import thread_schema
from .version import __version__